
USE16
STARTCODE
print_string:

;
;  Let ES:BP point to a null-terminated string
;
        jmp short init
        mov     ah,0Eh
        mov     bx,0            ;Set video page & Text olor
        SBEGIN
        DB 3EH,80H,7EH,00H,00H ;     cmp  BYTE PTR ds:[bp], 0
        SWHILE 0<>
         DB 3EH,8AH,46H,00H  ; mov     al, BYTE PTR ds:[bp]
          int     10h
          inc     bp
        SREPEAT
        ret
        DB 0
init:

;; Get extended memory size (kB)
        mov     ah, 88H
        int     15H
        mov WORD PTR [4000H], ax 
;Store the result somewhere for later
                                ;0x4000 is temporary..

;; Set keyboard repeat rate to 'fast'
        mov     ax, 0305H
        xor     bx, bx
        int     16H

;; Integrity check? (CRC? Checksum? Signature at end?)


;; Check for proper CPU type & mode
        call    check_cpu_type          ;Must be 386 or better
        call    check_mode              ;Must be in real mode

;; Check for FPU?


;; Get ready for PMODE
        call    build_GDT
        call    build_IDT
        call    reprogram_PICs
        call    enable_a20

;; ### Initialize PMODE ###
        cli
        DB 0FH,01H,16H,00H,0EH ; lgdt    [GDT_ADDRESS]
        DB 0FH,01H,1EH,69H,7FH ; lidt    [idtr]
        DB 0FH,20H,0C0H ; mov     eax, CR0
        inc     eax
        DB 0FH,22H,0C0H ; mov     CR0, eax
        DB 66H,0EAH,0C0H,7FH,00H,00H,08H,00H ;  jmp     dword 8:pmode

;-------------------------------------------------------------------------


;; Enable the A20 line so we can access memory above 1M
enable_a20:
        call    empty_8042
        mov     al, 0D1H        ;write command
        out     064H, al
        call    empty_8042
        mov     al, 0DFH        ;Enable A20
        out     60H, al
        call    empty_8042


;; Wait for A20 to actually turn on (on some computers, especially laptops,
;; it takes long enough to cause problems)
TEST_ADDR: EQU  7CH
        push    ds
        xor     ax, ax          ;ds = 0x0000
        mov     ds, ax
        dec     ax              ;gs = 0xffff
        mov     gs, ax
        mov     bx, [TEST_ADDR] ;save value
        SBEGIN
          inc   ax              ;ax = 0
          mov   [TEST_ADDR], ax
          DB 65H,3BH,06H,8CH,00H ; cmp   ax, gs: [TEST_ADDR+10H] ;test wrapped-around location
        SUNTIL 0<>                       ;wait until wrap-around stops
        mov     [TEST_ADDR], bx ;restore value
        pop     ds
        ret

;Flush the keyboard controller
empty_8042:
        SBEGIN
          SBEGIN
            jmp short $+2
            in      al, 64H        ;Status port
            test    al, 1           ;Output buffer?
          SWHILE  0<>
            jmp short $+2
            in      al, 60H        ;Read byte
          SREPEAT
          test    al, 2           ;Input buffer full?
        SUNTIL 0=
        ret


;Reprogram 8259A programmable interrupt controllers
;
;The BIOS puts interrupts at 0x08-0x0F, and we want them at 0x20-0x2F,
;right after the reserved interrupts, where they belong.
;
;Code snipped from John Fine's V86TEST
;
reprogram_PICs:

        M_PIC:  EQU     020H            ;I/O for master PIC
        M_IMR:  EQU     021H            ;I/O for master IMR
        S_PIC:  EQU     0A0H            ;I/O for slave PIC
        S_IMR:  EQU     0A1H            ;I/O for slace IMR

        EOI:    EQU     020H            ;EOI command

        ICW1:   EQU     011H            ;Cascade, Edge triggered
                                        ;ICW2 is vector
                                        ;ICW3 is slave bitmap or number
        ICW4:   EQU     001H            ;8088 mode

        M_VEC:  EQU     020H            ;Vector for master
        S_VEC:  EQU     028H            ;Vector for slave

        OCW3_IRR: EQU   00AH            ;Read IRR
        OCW3_ISR: EQU   00BH            ;Read ISR

        mov     al, ICW1                ;Start 8259 initialization
        out     M_PIC, al
        out     S_PIC, al
        mov     al, M_VEC               ;Base interrupt vector
        out     M_PIC+1, al
        mov     al, S_VEC
        out     S_PIC+1, al
        mov     al, 4                   ;Bitmask for cascade on IRQ 2
        out     M_PIC+1, al
        mov     al, 2                   ;Cascade on IRQ 2
        out     S_PIC+1, al
        mov     al, ICW4                ;Finish 8259 initialization
        out     M_PIC+1, al
        out     S_PIC+1, al

        mov     al, 0FFH                ;Mask all interrupts
        out     M_IMR, al
        out     S_IMR, al

        ret

;Check for 386, etc.
check_cpu_type:
        pushf
        xor     ah, ah
        push    ax
        popf
        pushf
        pop     ax
        and     ah, 0F0H
        cmp     ah, 0F0H
        SIF 0<> 
          mov   ah, 070H
          push  ax
        popf
        pushf
          pop   ax
          and   ah, 070H
          SIF 0<> 
            popf
            ret
          STHEN
        STHEN
        mov  bp, msg + { 0 >T }
        call    print_string
        jmp short $
msg:    DB  "You need at least a 386 to run in protected mode!",10,13,0


;Make sure we're starting from Real mode
check_mode:
        DB 0FH,20H,0C0H ; mov     eax, CR0
        and     al, 1
        SIF     0=   ; .not_real_mode
          ret
        STHEN
        mov     bp,  msg1 + { 0 >T }
        DB 0E8H,0E4H,0FEH ; call    print_string
        jmp short $
msg1:   db      "You're already in protected mode! WTF?",10,13,0

IDT_ADDRESS: EQU 600H
build_IDT:
;
; Build a full-size IDT (256 entries), with all of them handled by the
; "Unknown Interrupt" routine. Interrupt handlers will be expected to
; install themselves by modifying the IDT.
;
        mov     ax, 60H ; { IDT_ADDRESS 0x10 / }      ;ES:DI = start of IDT
        mov     es, ax
        xor     di, di
        mov     cx, 256                 ;CX = # of entries
        cld
loop1:
        push    cx

        mov     si, entry + { 0 >T }    ;DS:SI = template entry
        mov     cx, 4H
        rep movsw                   ;Copy it...

        pop     cx
        loop    loop1
        ret
entry:  dw      1387H
        dw      8
        db      0
        db      8EH
        dw      0


idtr:   ;For the IDT generated on the fly (by build_IDT)
        dw      800H ;  256*8
        dd      IDT_ADDRESS

        GDT_ADDRESS:    equ     E00H
build_GDT:
        mov     ax, 0E0H ; (GDT_ADDRESS >> 4)
        mov     es, ax
        xor     di, di
;        DB 0BEH,90H,7FH ;  
        mov     si, gdt + { 0 >T } 
        mov     cx, 18H ; (gdt_end - gdt) / 2
        cld
        rep movsw

        xor     di, di
        DB  26H,0C7H,45H,02H,00H,0EH ; mov   es:[di+2], GDT_ADDRESS
        DB  26H,0C7H,45H,04H,00H,00H ; mov  WORD PTR es:[di+4], 0
        ret


        DB  90H, 90H ; ALIGN 4
;General Descriptor Table
gdt:
        DW  30H      ;  %$limitoftable
        DD  7F90H    ;  %$startoftable
        DW  0

;flat_code
     DW  0FFFFH       ; 4Gb - (0x100000*0x1000 = 4Gb)
     DW   0000H       ; base address=0
     DW   9A00H       ; code read/exec
     DW   00CFH       ; granularity=4096, 386 (+5th nibble of limit)
;flat_data
     DW  0FFFFH       ; 4Gb - (0x100000*0x1000 = 4Gb)
     DW   0000H       ; base address=0
     DW   9200H       ; data read/write
     DW   00CFH       ; granularity=4096, 386 (+5th nibble of limit)
;user_code
DD  0000FFFFH, 00CFFA00H ;  0, 0xFFFFF, D_CODE+D_READ+D_BIG+D_BIG_LIM+D_DPL3
user_data:
DD  0000FFFFH, 00CFF200H ;  0, 0xFFFFF, D_DATA+D_WRITE+D_BIG+D_BIG_LIM+D_DPL3
tss_sel:
DD  00000000H, 00008900H ;  0, 0, D_TSS

EndCode

USE32
STARTCODE

;
; Copyright (C) 1999 Thomas Novelli
;       Retro is distributed under the terms of the GNU LGPL, version 2 or
;       greater. (For the text of the license, see the file COPYING.LIB)
;

RETURN_STACK:    EQU 00210000H
FORTH_STACK:     EQU 00220000H

pmode:
        mov     ax, 10H ; flat_data
        mov     ds, ax
        mov     es, ax
        mov     fs, ax
        mov     gs, ax
        mov     ss, ax
        mov     esp, RETURN_STACK
        mov     ebp, FORTH_STACK

        mov     al, 11111011b   ;disable all IRQs except cascade (IRQ 2)
        out      21H, al
        mov     al, 11111111b
        out     0A1H, al
        sti
EndCode
